from pandas_datareader.macro.eurostat import EurostatClient
from pandas_datareader.macro.oecd import OECDClient


def _get_client(provider):
    if provider == "oecd":
        return OECDClient()
    if provider == "eurostat":
        return EurostatClient()
    raise ValueError(f"Unknown provider: {provider}")


def read_macro(provider, dataset, **kwargs):
    return _get_client(provider).read(dataset, **kwargs)


def search_macro_datasets(provider, query=None, **kwargs):
    return _get_client(provider).search_datasets(query=query, **kwargs)


def describe_macro_dataset(provider, dataset, **kwargs):
    return _get_client(provider).describe_dataset(dataset, **kwargs)



"""Sphinx configuration for the pandas-datareader documentation."""

from __future__ import annotations

from datetime import date
from pathlib import Path

import pandas_datareader as pdr

SOURCE_DIR = Path(__file__).parent

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

project = "pandas-datareader"
author = "The PyData Development Team"
copyright = f"{date.today():%Y}, {author}"

release = pdr.__version__
version = release.split("+", 1)[0]
if ".dev" in version and "+" in release:
    commits_since_tag = version.rsplit(".dev", 1)[1]
    commit_hash = release.split("+", 1)[1]
    version = f"{version} (+{commits_since_tag}, {commit_hash})"

with (SOURCE_DIR / "_version.txt").open("w", encoding="utf-8") as version_file:
    doc_date = date.today().strftime("%B %d, %Y")
    version_file.write(f"Version: **{version}** Date: **{doc_date}**\n")

root_doc = "index"
master_doc = root_doc
source_suffix = {".rst": "restructuredtext"}
language = "en"
templates_path = ["_templates"] if (SOURCE_DIR / "_templates").exists() else []
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "default"
todo_include_todos = True

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

extlinks = {
    "issue": ("https://github.com/pydata/pandas-datareader/issues/%s", "GH %s"),
    "wiki": ("https://github.com/pydata/pandas-datareader/wiki/%s", "wiki %s"),
}


# -- Options for HTML output ----------------------------------------------

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "github_url": "https://github.com/pydata/pandas-datareader",
    "logo": {
        "alt_text": "pandas-datareader",
        "image_light": "_static/images/pandas-datareader-plain.svg",
        "image_dark": "_static/images/pandas-datareader-plain-dark.svg",
    },
    "navigation_with_keys": True,
}
html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = "pandas-datareaderdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}
latex_documents = [
    (
        root_doc,
        "pandas-datareader.tex",
        "pandas-datareader Documentation",
        "pydata",
        "manual",
    )
]


# -- Options for manual page output ---------------------------------------

man_pages = [
    (root_doc, "pandas-datareader", "pandas-datareader Documentation", [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (
        root_doc,
        "pandas-datareader",
        "pandas-datareader Documentation",
        author,
        "pandas-datareader",
        "One line description of project.",
        "Miscellaneous",
    )
]



from pandas import concat, read_csv

from pandas_datareader.base import _BaseReader
from pandas_datareader.compat import is_list_like


class FredReader(_BaseReader):
    """
    Get data for the given name from the St. Louis FED (FRED).
    """

    @property
    def url(self):
        """API URL"""
        return "https://fred.stlouisfed.org/graph/fredgraph.csv"

    def read(self):
        """Read data

        Returns
        -------
        data : DataFrame
            If multiple names are passed for "series" then the index of the
            DataFrame is the outer join of the indices of each series.
        """
        try:
            return self._read()
        finally:
            self.close()

    def _read(self):
        if not is_list_like(self.symbols):
            names = [self.symbols]
        else:
            names = self.symbols

        urls = [f"{self.url}?id={n}" for n in names]

        def fetch_data(url, name):
            """Utility to fetch data"""
            resp = self._read_url_as_StringIO(url)
            data = read_csv(
                resp,
                index_col=0,
                parse_dates=True,
                header=None,
                skiprows=1,
                names=["DATE", name],
                na_values=".",
            )
            try:
                return data.truncate(self.start, self.end)
            except KeyError as exc:  # pragma: no cover
                if data.iloc[3].name[7:12] == "Error":
                    raise OSError(
                        "Failed to get the data. Check that "
                        "{!r} is a valid FRED series.".format(name)
                    ) from exc
                raise

        try:
            data = [fetch_data(url, n) for url, n in zip(urls, names, strict=True)]
        except TypeError:
            # Python 3.9 only
            data = [fetch_data(url, n) for url, n in zip(urls, names)]  # noqa: B905
        df = concat(
            data,
            axis=1,
            join="outer",
        )
        return df



"""
Get a list of contributors between now and a start date
"""

import datetime as dt
import os

from github import Github

# or using an access token
g = Github(os.environ.get("GITHUB_API_KEY"))

repo = g.get_repo("pydata/pandas-datareader")

start = dt.datetime(2019, 9, 26, 0, 0, 0)
pulls = repo.get_pulls(state="closed", sort="updated", base="master", direction="desc")

users = set()
for i, p in enumerate(pulls):
    print(f"{i}: {p.number}")
    if p.merged and p.merged_at >= start:
        users.update([p.user.name])
    if p.updated_at < start:
        break

contrib = sorted((c for c in users if c), key=lambda s: s.split(" ")[-1])

for c in contrib:
    print(f"- {c}")



import datetime as dt

from pandas import to_datetime
import requests

from pandas_datareader.compat import is_number


class SymbolWarning(UserWarning):
    pass


class RemoteDataError(IOError):
    pass


def _sanitize_dates(start, end):
    """
    Return (timestamp_start, timestamp_end) tuple
    if start is None - default is 5 years before the current date
    if end is None - default is today

    Parameters
    ----------
    start : str, int, date, datetime, Timestamp
        Desired start date
    end : str, int, date, datetime, Timestamp
        Desired end date
    """
    if start is None:
        # default to 5 years before today
        today = dt.date.today()
        start = today - dt.timedelta(days=365 * 5)
    elif is_number(start):
        # regard int as year
        start = dt.datetime(start, 1, 1)

    if end is None:
        # default to today
        end = dt.date.today()
    elif is_number(end):
        end = dt.datetime(end, 1, 1)

    try:
        start = to_datetime(start)
        end = to_datetime(end)
    except (TypeError, ValueError) as exc:
        raise ValueError("Invalid date format.") from exc
    if start > end:
        raise ValueError("start must be an earlier date than end")
    return start, end


def _init_session(session):
    if session is None:
        session = requests.Session()
        # do not set requests max_retries here to support arbitrary pause
    else:
        if not isinstance(session, requests.Session):
            raise TypeError("session must be a request.Session")
    return session



import pandas as pd

from pandas_datareader.eurostat import EurostatReader


def test_api_urls():
    reader = EurostatReader(
        "ert_h_eur_a",
        start=pd.Timestamp("2009-01-01"),
        end=pd.Timestamp("2010-01-01"),
    )

    assert reader.url.startswith(
        "https://ec.europa.eu/eurostat/api/dissemination/sdmx/2.1/data/"
    )
    assert reader.dsd_url.startswith(
        "https://ec.europa.eu/eurostat/api/dissemination/sdmx/2.1/datastructure/"
    )


def test_parse_statistics_payload():
    payload = {
        "id": ["geo", "time"],
        "size": [2, 2],
        "dimension": {
            "geo": {
                "label": "Geopolitical entity",
                "category": {
                    "index": {"DE": 0, "FR": 1},
                    "label": {"DE": "Germany", "FR": "France"},
                },
            },
            "time": {
                "label": "Time period",
                "category": {
                    "index": {"2020": 0, "2021": 1},
                    "label": {"2020": "2020", "2021": "2021"},
                },
            },
        },
        "value": {"0": 1.0, "1": 2.0, "2": 3.0, "3": 4.0},
    }

    result = EurostatReader._read_statistics_payload(payload)

    expected = pd.DataFrame(
        [[1.0, 3.0], [2.0, 4.0]],
        index=pd.DatetimeIndex(["2020-01-01", "2021-01-01"], name="TIME_PERIOD"),
        columns=pd.Index(["Germany", "France"], name="GEO"),
    )

    pd.testing.assert_frame_equal(result, expected)



from datetime import datetime

import numpy as np
import pandas as pd
from pandas import DataFrame, testing as tm
import pytest

from pandas_datareader import data as web
from pandas_datareader._utils import RemoteDataError

pytestmark = pytest.mark.stable


class TestFred:
    def test_fred(self):
        # Raises an exception when DataReader can't
        # get a 200 response from FRED.

        start = datetime(2010, 1, 1)
        end = datetime(2013, 1, 1)

        df = web.DataReader("GDP", "fred", start, end)
        ts = df["GDP"]

        assert ts.index[0] == pd.to_datetime("2010-01-01")
        assert ts.index[-1] == pd.to_datetime("2013-01-01")
        assert ts.index.name == "DATE"
        assert ts.name == "GDP"
        assert len(ts) == 13

        with pytest.raises(RemoteDataError):
            web.DataReader("NON EXISTENT SERIES", "fred", start, end)

    def test_fred_nan(self):
        start = datetime(2010, 1, 1)
        end = datetime(2013, 1, 27)
        df = web.DataReader("DFII5", "fred", start, end)
        assert pd.isnull(df.loc["2010-01-01"].iloc[0])

    def test_fred_parts(self):  # pragma: no cover
        start = datetime(2010, 1, 1)
        end = datetime(2013, 1, 27)
        df = web.get_data_fred("CPIAUCSL", start, end)
        assert df.loc["2010-05-01"].iloc[0] == 217.29

        t = df.CPIAUCSL.values
        assert np.issubdtype(t.dtype, np.floating)
        assert t.shape == (37,)

    def test_fred_part2(self):
        expected = [[576.7], [962.9], [684.7], [848.3], [933.3]]
        result = web.get_data_fred("A09024USA144NNBR", start="1915").iloc[:5]
        np.testing.assert_array_equal(result.values, np.array(expected))

    def test_invalid_series(self):
        name = "NOT A REAL SERIES"
        with pytest.raises(Exception):  # noqa: B017
            web.get_data_fred(name)

    def test_fred_multi(self):  # pragma: no cover
        names = ["CPIAUCSL", "CPALTT01USQ661S", "CPILFESL"]
        start = datetime(2010, 1, 1)
        end = datetime(2013, 1, 27)

        received = web.DataReader(names, "fred", start, end).head(1)

        expected = DataFrame(
            [[217.488, 91.712409, 220.633]],
            columns=names,
            index=[pd.Timestamp("2010-01-01 00:00:00")],
        )
        expected.index.rename("DATE", inplace=True)
        expected.index.freq = "MS"
        np.testing.assert_allclose(received, expected)
        tm.assert_index_equal(received.index, expected.index)
        tm.assert_index_equal(received.columns, expected.columns)

    def test_fred_multi_bad_series(self):
        names = ["NOTAREALSERIES", "CPIAUCSL", "ALSO FAKE"]
        with pytest.raises(RemoteDataError):
            web.DataReader(names, data_source="fred")



import datetime as dt

import pandas as pd
import pytest

from pandas_datareader._utils import _sanitize_dates


class TestUtils:
    @pytest.mark.parametrize(
        "input_date",
        [
            "2019-01-01",
            "JAN-01-2010",
            dt.datetime(2019, 1, 1),
            dt.date(2019, 1, 1),
            pd.Timestamp(2019, 1, 1),
        ],
    )
    def test_sanitize_dates(self, input_date):
        expected_start = pd.to_datetime(input_date)
        expected_end = pd.to_datetime(dt.date.today())
        result = _sanitize_dates(input_date, None)
        assert result == (expected_start, expected_end)

    def test_sanitize_dates_int(self):
        start_int = 2018
        end_int = 2019
        expected_start = pd.to_datetime(dt.datetime(start_int, 1, 1))
        expected_end = pd.to_datetime(dt.datetime(end_int, 1, 1))
        assert _sanitize_dates(start_int, end_int) == (expected_start, expected_end)

    def test_sanitize_invalid_dates(self):
        with pytest.raises(ValueError):
            _sanitize_dates(2019, 2018)

        with pytest.raises(ValueError):
            _sanitize_dates("2019-01-01", "2018-01-01")

        with pytest.raises(ValueError):
            _sanitize_dates("20199", None)

    def test_sanitize_dates_defaults(self):
        default_start = pd.to_datetime(dt.date.today() - dt.timedelta(days=365 * 5))
        default_end = pd.to_datetime(dt.date.today())
        assert _sanitize_dates(None, None) == (default_start, default_end)



import os

from pandas_datareader.compat import get_filepath_or_buffer


def _read_content(path_or_buf):
    """
    Copied part of internal logic from pandas.io.read_json.
    """

    filepath_or_buffer = get_filepath_or_buffer(path_or_buf)[0]

    if isinstance(filepath_or_buffer, str):
        try:
            exists = os.path.exists(filepath_or_buffer)
        except (TypeError, ValueError):
            exists = False

        if exists:
            with open(filepath_or_buffer) as fh:
                data = fh.read()
        else:
            data = filepath_or_buffer
    elif hasattr(filepath_or_buffer, "read"):
        data = filepath_or_buffer.read()
    else:
        data = filepath_or_buffer

    return data



from datetime import date, timedelta

import pytest

from pandas_datareader import data as web
from pandas_datareader._utils import RemoteDataError

pytestmark = pytest.mark.stable


class TestBankOfCanada:
    @staticmethod
    def get_symbol(currency_code, inverted=False):
        if inverted:
            return f"FXCAD{currency_code}"
        else:
            return f"FX{currency_code}CAD"

    def check_bankofcanada_count(self, code):
        start, end = date.today() - timedelta(days=30), date.today()
        df = web.DataReader(self.get_symbol(code), "bankofcanada", start, end)
        assert df.size > 15

    def check_bankofcanada_valid(self, code):
        symbol = self.get_symbol(code)
        df = web.DataReader(
            symbol, "bankofcanada", date.today() - timedelta(days=30), date.today()
        )
        assert symbol in df.columns

    def check_bankofcanada_inverted(self, code):
        symbol = self.get_symbol(code)
        symbol_inverted = self.get_symbol(code, inverted=True)

        df = web.DataReader(
            symbol, "bankofcanada", date.today() - timedelta(days=30), date.today()
        )
        df_i = web.DataReader(
            symbol_inverted,
            "bankofcanada",
            date.today() - timedelta(days=30),
            date.today(),
        )
        try:
            pairs = zip(
                (1 / df)[symbol].tolist(), df_i[symbol_inverted].tolist(), strict=True
            )
        except TypeError:
            # Python 3.9 only
            pairs = zip(  # noqa: B905
                (1 / df)[symbol].tolist(), df_i[symbol_inverted].tolist()
            )
        assert all(a - b < 0.01 for a, b in pairs)

    def test_bankofcanada_usd_count(self):
        self.check_bankofcanada_count("USD")

    def test_bankofcanada_eur_count(self):
        self.check_bankofcanada_count("EUR")

    def test_bankofcanada_usd_valid(self):
        self.check_bankofcanada_valid("USD")

    def test_bankofcanada_eur_valid(self):
        self.check_bankofcanada_valid("EUR")

    def test_bankofcanada_usd_inverted(self):
        self.check_bankofcanada_inverted("USD")

    def test_bankofcanada_eur_inverted(self):
        self.check_bankofcanada_inverted("EUR")

    def test_bankofcanada_bad_range(self):
        with pytest.raises(ValueError):
            web.DataReader(
                "FXCADUSD",
                "bankofcanada",
                date.today(),
                date.today() - timedelta(days=30),
            )

    def test_bankofcanada_bad_url(self):
        with pytest.raises(RemoteDataError):
            web.DataReader(
                "abcdefgh",
                "bankofcanada",
                date.today() - timedelta(days=30),
                date.today(),
            )



from datetime import datetime

import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader import data as web
from pandas_datareader._utils import RemoteDataError


class TestOECD:
    def test_get_un_den(self):
        df = web.DataReader(
            "TUD", "oecd", start=datetime(1960, 1, 1), end=datetime(2012, 1, 1)
        )
        japan = df["Japan"].iloc[:, 0]
        united_states = df["United States"].iloc[:, 0]

        expected_index = pd.date_range(
            "1960-01-01", "1964-01-01", freq="YS", name="Time period"
        )
        expected = pd.Series(
            [24.299999, 24.299999, 24.0, 23.4, 22.200001],
            index=expected_index,
            name=("Trade union density", "Percentage of employees"),
        )
        tm.assert_series_equal(japan.dropna().iloc[:5], expected, check_freq=False)
        if len(united_states.dropna()) > 0:
            assert united_states.first_valid_index() == pd.Timestamp("1982-01-01")
            assert united_states.last_valid_index() == pd.Timestamp("2002-01-01")
            assert int(united_states.notna().sum()) == 19
        else:
            raise AssertionError("United States data is missing in the OECD dataset.")

    def test_get_tourism(self):
        df = web.DataReader(
            "TOURISM_INBOUND",
            "oecd",
            start=datetime(2008, 1, 1),
            end=datetime(2012, 1, 1),
        )

        jp = np.array([8351000, 6790000, 8611000, 6219000, 8368000], dtype=float)
        us = np.array(
            [175702309, 160507417, 164079732, 167600277, 171320408], dtype=float
        )
        index = pd.date_range("2008-01-01", "2012-01-01", freq="YS", name="Time period")
        current_name = (
            "Not applicable",
            "Arrivals",
            "Persons",
            "All visitors (overnight and same-day)",
            "Not applicable",
            "Tourism demand surveys",
        )
        for label, values in [("Japan", jp), ("United States", us)]:
            expected = pd.Series(values, index=index, name="Tourism demand surveys")
            expected.index.freq = None
            series = df[label][current_name].loc["2008":"2012"].astype(float)
            np.testing.assert_allclose(series.to_numpy(), values, rtol=0, atol=10000)
            tm.assert_index_equal(series.index, expected.index)

    def test_oecd_invalid_symbol(self):
        with pytest.raises(RemoteDataError):
            web.DataReader("INVALID_KEY", "oecd")

        with pytest.raises(ValueError):
            web.DataReader(1234, "oecd")



from __future__ import annotations

from xml.etree import ElementTree as ET

import pandas as pd

from pandas_datareader.io.sdmx import _read_sdmx_dsd, read_sdmx
from pandas_datareader.macro.base import MacroClientBase
from pandas_datareader.macro.result import MacroResult

_MESSAGE = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message}"
_STRUCTURE = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure}"
_COMMON = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common}"
_XML = "{http://www.w3.org/XML/1998/namespace}"


def _preferred_text(parent, tag):
    nodes = list(parent.findall(tag))
    if not nodes:
        return None
    for node in nodes:
        lang = node.get(f"{_XML}lang", "")
        if lang.lower().startswith("en"):
            return node.text
    return nodes[0].text


class EurostatClient(MacroClientBase):
    provider = "eurostat"
    base_url = "https://ec.europa.eu/eurostat/api/dissemination/sdmx/2.1"

    def _parse_dataflows_xml(self, xml_bytes):
        root = ET.fromstring(xml_bytes)
        rows = []
        for flow in root.iter(_STRUCTURE + "Dataflow"):
            if flow.get("agencyID") != "ESTAT":
                continue
            dataflow_id = flow.get("id")
            version = flow.get("version")
            structure_ref = flow.find(f".//{_STRUCTURE}Structure/Ref")
            rows.append(
                {
                    "dataset_id": dataflow_id.lower(),
                    "dataflow_id": dataflow_id,
                    "agency_id": "ESTAT",
                    "version": version,
                    "title": _preferred_text(flow, f"{_COMMON}Name"),
                    "description": _preferred_text(flow, f"{_COMMON}Description"),
                    "structure_id": (
                        None if structure_ref is None else structure_ref.get("id")
                    ),
                }
            )
        return pd.DataFrame(rows)

    def _parse_dsd_xml(self, xml_bytes):
        root = ET.fromstring(xml_bytes)
        codelists = {}
        for codelist in root.iter(_STRUCTURE + "Codelist"):
            entries = {}
            for code in codelist.iter(_STRUCTURE + "Code"):
                entries[code.get("id")] = _preferred_text(code, f"{_COMMON}Name")
            codelists[codelist.get("id")] = entries
        return root, codelists

    def _parse_dsd_metadata(self, xml_bytes):
        root, codelists = self._parse_dsd_xml(xml_bytes)
        dimensions = []
        labels = {}
        for dim in root.findall(
            f".//{_STRUCTURE}DataStructure//{_STRUCTURE}DimensionList/*"
        ):
            dim_id = dim.get("id")
            enum_ref = dim.find(f".//{_STRUCTURE}Enumeration/Ref")
            values = {}
            if enum_ref is not None:
                values = codelists.get(enum_ref.get("id"), {})
            dimensions.append({"id": dim_id, "label": dim_id, "values": values})
            labels[dim_id] = values
        return {"dimensions": dimensions, "labels": labels, "codelists": codelists}

    @staticmethod
    def _format_columns(data):
        if not isinstance(data.columns, pd.MultiIndex):
            return data

        columns = data.columns
        order_priority = {
            "currency": 0,
            "statinfo": 1,
            "unit": 2,
            "siec": 3,
            "tax": 4,
            "nrg_cons": 5,
            "geo": 98,
            "freq": 99,
        }
        current = list(columns.names)
        order = sorted(
            range(len(current)), key=lambda i: (order_priority.get(current[i], 50), i)
        )
        columns = columns.reorder_levels(order)
        columns = columns.set_names([name.upper() for name in columns.names])
        data.columns = columns
        return data

    def _build_result_from_payload(self, dataset_id, data_xml, dsd_xml, query=None):
        dsd = _read_sdmx_dsd(dsd_xml)
        metadata = self._parse_dsd_metadata(dsd_xml)
        data = read_sdmx(data_xml, dsd=dsd)
        data = self._format_columns(data)

        root = ET.fromstring(data_xml)
        prepared = root.find(f".//{_MESSAGE}Prepared")
        result_metadata = {
            "title": dataset_id,
            "description": dataset_id,
            "updated_at": None if prepared is None else prepared.text,
            "dimensions": metadata["dimensions"],
            "attributes": {},
            "labels": metadata["labels"],
            "source": "eurostat-sdmx-2.1",
            "query": query or {},
            "notes": [],
            "raw": {"dataset_id": dataset_id},
        }
        return MacroResult(
            data=data,
            metadata=result_metadata,
            provider=self.provider,
            dataset_id=dataset_id,
        )

    def _dataflow_url(self, dataset=None):
        if dataset is None:
            return f"{self.base_url}/dataflow/all/all/latest"
        return f"{self.base_url}/dataflow/ESTAT/{dataset.upper()}/latest"

    def _dsd_url(self, dataset):
        return f"{self.base_url}/datastructure/ESTAT/{dataset}/latest?references=descendants"

    def _data_url(self, dataset, start=None, end=None):
        url = f"{self.base_url}/data/{dataset}"
        params = []
        if start is not None:
            params.append(f"startPeriod={pd.to_datetime(start).year}")
        if end is not None:
            params.append(f"endPeriod={pd.to_datetime(end).year}")
        if params:
            url = f"{url}?{'&'.join(params)}"
        return url

    def search_datasets(self, query=None, **kwargs):
        response = self._get(self._dataflow_url())
        result = self._parse_dataflows_xml(response.content)
        if query:
            q = query.lower()
            mask = (
                result["dataset_id"].fillna("").str.lower().str.contains(q)
                | result["title"].fillna("").str.lower().str.contains(q)
                | result["description"].fillna("").str.lower().str.contains(q)
            )
            result = result.loc[mask].reset_index(drop=True)
        return result

    def describe_dataset(self, dataset, **kwargs):
        dsd_xml = self._get(self._dsd_url(dataset)).content
        dataflow_xml = self._get(self._dataflow_url(dataset)).content
        dataflow = self._parse_dataflows_xml(dataflow_xml).iloc[0].to_dict()
        dsd_meta = self._parse_dsd_metadata(dsd_xml)
        return {
            "dataset_id": dataset,
            "title": dataflow["title"],
            "description": dataflow["description"],
            "updated_at": None,
            "dimensions": dsd_meta["dimensions"],
            "attributes": {},
            "labels": dsd_meta["labels"],
            "source": "eurostat-sdmx-2.1",
            "query": {},
            "notes": [],
            "raw": {"dataflow": dataflow},
        }

    def read(self, dataset, start=None, end=None, filters=None, **kwargs):
        if filters is not None:
            raise NotImplementedError(
                "Eurostat dict/string filters are not implemented in the new macro API yet."
            )
        data_xml = self._get(self._data_url(dataset, start=start, end=end)).content
        dsd_xml = self._get(self._dsd_url(dataset)).content
        result = self._build_result_from_payload(
            dataset,
            data_xml,
            dsd_xml,
            query={"start": start, "end": end, "filters": filters},
        )
        described = self.describe_dataset(dataset)
        result.metadata["title"] = described["title"]
        result.metadata["description"] = described["description"]
        return result



import numpy as np
import pandas as pd
from pandas import testing as tm

from pandas_datareader import data as web
from pandas_datareader._testing import skip_on_exception
from pandas_datareader._utils import RemoteDataError


class TestEurostat:
    def test_get_ert_h_eur_a(self):
        # Former euro area national currencies vs. euro/ECU
        # annual data (ert_h_eur_a)
        df = web.DataReader(
            "ert_h_eur_a",
            "eurostat",
            start=pd.Timestamp("2009-01-01"),
            end=pd.Timestamp("2010-01-01"),
        )
        assert isinstance(df, pd.DataFrame)

        currencies = ["Italian lira", "Lithuanian litas"]
        # cols = [(currency, "Average") for currency in currencies]
        df_currency = df[currencies]
        df_currency = df_currency.xs("Average", axis=1, level=1)
        df_currency.columns = df_currency.columns.droplevel(1)

        exp_col = pd.MultiIndex.from_product(
            [currencies, ["Annual"]], names=["CURRENCY", "FREQ"]
        )
        exp_idx = pd.DatetimeIndex(["2009-01-01", "2010-01-01"], name="TIME_PERIOD")
        values = np.array([[1936.27, 3.4528], [1936.27, 3.4528]])
        expected = pd.DataFrame(values, index=exp_idx, columns=exp_col)
        tm.assert_frame_equal(df_currency, expected)

    def test_get_sts_cobp_a(self):
        # Building permits - annual data (2010 = 100)
        df = web.DataReader(
            "sts_cobp_a",
            "eurostat",
            start=pd.Timestamp("2000-01-01"),
            end=pd.Timestamp("2013-01-01"),
        )

        idx = pd.date_range("2000-01-01", "2013-01-01", freq="YS", name="TIME_PERIOD")
        ne_name = (
            "Building permits - m2 of useful floor area",
            "Index, 2015=100",
            "Unadjusted data (i.e. neither seasonally adjusted nor "
            "calendar adjusted data)",
            "Non-residential buildings, except office buildings",
            "Netherlands",
            "Annual",
        )
        ne_values = [
            258.4,
            240.9,
            164.5,
            168.8,
            185.1,
            190.9,
            228.3,
            293.7,
            257.7,
            166.0,
            129.2,
            147.0,
            118.4,
            98.0,
        ]
        ne = pd.Series(ne_values, name=ne_name, index=idx)

        uk_name = (
            "Building permits - m2 of useful floor area",
            "Index, 2015=100",
            "Unadjusted data (i.e. neither seasonally adjusted nor "
            "calendar adjusted data)",
            "Non-residential buildings, except office buildings",
            "United Kingdom",
            "Annual",
        )
        uk_values = [
            137.0,
            137.9,
            134.1,
            136.5,
            144.9,
            137.1,
            137.6,
            148.3,
            138.7,
            128.9,
            121.7,
            120.0,
            126.2,
            99.0,
        ]
        uk = pd.Series(uk_values, name=uk_name, index=idx)

        for expected in [ne, uk]:
            expected.index.freq = None
            result = df[expected.name]
            tm.assert_series_equal(result, expected, check_freq=False)

    def test_get_nrg_pc_202(self):
        # see gh-149

        df = web.DataReader(
            "nrg_pc_202",
            "eurostat",
            start=pd.Timestamp("2010-01-01"),
            end=pd.Timestamp("2013-01-01"),
        )

        name = (
            "Euro",
            "Gigajoule (gross calorific value - GCV)",
            "Natural gas",
            "All taxes and levies included",
            "Consumption less than 20 GJ - band D1",
            "Denmark",
            "Half-yearly, semesterly",
        )

        exp_index = pd.Index(
            [
                "2010-S1",
                "2010-S2",
                "2011-S1",
                "2011-S2",
                "2012-S1",
                "2012-S2",
                "2013-S1",
                "2013-S2",
            ],
            name="TIME_PERIOD",
        )
        exp = pd.Series(
            [26.4979, 26.8147, 28.6448, 28.5862, 27.2187, 26.5285, 27.5854, 27.1403],
            index=exp_index,
            name=name,
        )

        tm.assert_series_equal(df[name], exp)

    @skip_on_exception(RemoteDataError)
    def test_get_prc_hicp_manr_large_query(self):
        df = web.DataReader(
            "prc_hicp_manr",
            "eurostat",
            start=pd.Timestamp("2010-01-01"),
            end=pd.Timestamp("2013-01-01"),
        )
        assert isinstance(df, pd.DataFrame)
        assert isinstance(df.index, pd.DatetimeIndex)
        assert df.index.min() >= pd.Timestamp("2010-01-01")
        assert df.index.max() <= pd.Timestamp("2013-12-31")
        assert "FREQ" in df.columns.names
        assert "UNIT" in df.columns.names
        assert "COICOP" in df.columns.names
        assert "GEO" in df.columns.names



from __future__ import annotations

from xml.etree import ElementTree as ET

import pandas as pd

from pandas_datareader.io.jsdmx import read_jsdmx
from pandas_datareader.macro.base import (
    MacroClientBase,
    MacroNotFoundError,
    MacroSchemaError,
)
from pandas_datareader.macro.result import MacroResult

_MESSAGE = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message}"
_STRUCTURE = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure}"
_COMMON = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common}"
_XML = "{http://www.w3.org/XML/1998/namespace}"


def _preferred_text(parent, tag):
    nodes = list(parent.findall(tag))
    if not nodes:
        return None
    for node in nodes:
        lang = node.get(f"{_XML}lang", "")
        if lang.lower().startswith("en"):
            return node.text
    return nodes[0].text


class OECDClient(MacroClientBase):
    provider = "oecd"
    dataflow_url = "https://sdmx.oecd.org/public/rest/dataflow/all/all/latest"
    data_url_template = (
        "https://sdmx.oecd.org/public/rest/data/{dataset}/{key}?format=jsondata"
    )

    def _parse_dataflows_xml(self, xml_bytes):
        root = ET.fromstring(xml_bytes)
        rows = []
        for flow in root.iter(_STRUCTURE + "Dataflow"):
            dataflow_id = flow.get("id")
            agency_id = flow.get("agencyID")
            version = flow.get("version")
            structure_ref = flow.find(f".//{_STRUCTURE}Structure/Ref")
            rows.append(
                {
                    "dataset_id": f"{agency_id},{dataflow_id},{version}",
                    "dataflow_id": dataflow_id,
                    "agency_id": agency_id,
                    "version": version,
                    "title": _preferred_text(flow, f"{_COMMON}Name"),
                    "description": _preferred_text(flow, f"{_COMMON}Description"),
                    "structure_id": (
                        None if structure_ref is None else structure_ref.get("id")
                    ),
                }
            )
        return pd.DataFrame(rows)

    @staticmethod
    def _extract_dimensions(payload):
        structure = payload["data"]["structures"][0]
        dimensions = []
        labels = {}
        for group_name, dims in structure.get("dimensions", {}).items():
            for dim in dims:
                values = {value["id"]: value["name"] for value in dim.get("values", [])}
                dimensions.append(
                    {
                        "group": group_name,
                        "id": dim.get("id"),
                        "label": dim.get("name"),
                        "values": values,
                    }
                )
                labels[dim.get("id")] = values
        return dimensions, labels

    def _build_result_from_payload(self, dataset_id, payload, query=None):
        data = read_jsdmx(payload)
        dimensions, labels = self._extract_dimensions(payload)
        structure = payload["data"]["structures"][0]
        metadata = {
            "title": structure.get("name"),
            "description": structure.get("description"),
            "updated_at": payload.get("meta", {}).get("prepared"),
            "dimensions": dimensions,
            "attributes": structure.get("attributes", {}),
            "labels": labels,
            "source": "oecd-sdmx-json-v2",
            "query": query or {},
            "notes": [],
            "raw": {
                "meta": payload.get("meta", {}),
                "errors": payload.get("errors", []),
            },
        }
        return MacroResult(
            data=data,
            metadata=metadata,
            provider=self.provider,
            dataset_id=dataset_id,
        )

    def _resolve_dataset(self, dataset):
        if "," in dataset:
            return dataset
        dataflows = self.search_datasets()
        exact = dataflows[dataflows["dataflow_id"] == dataset]
        if len(exact) == 1:
            return exact.iloc[0]["dataset_id"]
        raise MacroNotFoundError(f"Could not resolve OECD dataset: {dataset}")

    def _dataset_url(self, dataset_id, key="all", start=None, end=None):
        url = self.data_url_template.format(dataset=dataset_id, key=key)
        params = []
        if start is not None:
            params.append(f"startPeriod={self._normalize_date(start)}")
        if end is not None:
            params.append(f"endPeriod={self._normalize_date(end)}")
        if params:
            url = f"{url}&{'&'.join(params)}"
        return url

    def search_datasets(self, query=None, **kwargs):
        response = self._get(self.dataflow_url)
        result = self._parse_dataflows_xml(response.content)
        if query:
            q = query.lower()
            mask = (
                result["dataflow_id"].fillna("").str.lower().str.contains(q)
                | result["title"].fillna("").str.lower().str.contains(q)
                | result["description"].fillna("").str.lower().str.contains(q)
            )
            result = result.loc[mask].reset_index(drop=True)
        return result

    def describe_dataset(self, dataset, **kwargs):
        dataset_id = self._resolve_dataset(dataset)
        payload = self._get(
            f"{self._dataset_url(dataset_id)}&firstNObservations=1"
        ).json()
        result = self._build_result_from_payload(
            dataset_id,
            payload,
            query={"firstNObservations": 1},
        )
        return {
            "dataset_id": dataset_id,
            "title": result.metadata["title"],
            "description": result.metadata["description"],
            "updated_at": result.metadata["updated_at"],
            "dimensions": result.metadata["dimensions"],
            "attributes": result.metadata["attributes"],
            "labels": result.metadata["labels"],
            "source": result.metadata["source"],
            "query": result.metadata["query"],
            "notes": result.metadata["notes"],
            "raw": result.metadata["raw"],
        }

    def read(self, dataset, start=None, end=None, filters=None, **kwargs):
        dataset_id = self._resolve_dataset(dataset)
        key = "all" if filters is None else filters
        if not isinstance(key, str):
            raise MacroSchemaError(
                "OECD filters must be a provider-specific key string."
            )
        payload = self._get(
            self._dataset_url(dataset_id, key=key, start=start, end=end)
        ).json()
        return self._build_result_from_payload(
            dataset_id,
            payload,
            query={"start": start, "end": end, "filters": filters},
        )



import subprocess
import sys


def test_import_pandas_datareader_smoke():
    cmd = [
        sys.executable,
        "-c",
        "import pandas; import pandas_datareader; print(pandas.__version__)",
    ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    assert proc.returncode == 0, proc.stderr



"""
Module contains tools for collecting data from various remote sources
"""

# flake8: noqa

import warnings

from pandas_datareader.bankofcanada import BankOfCanadaReader
from pandas_datareader.compat import deprecate_kwarg
from pandas_datareader.econdb import EcondbReader
from pandas_datareader.eurostat import EurostatReader
from pandas_datareader.exceptions import DEP_ERROR_MSG, ImmediateDeprecationError
from pandas_datareader.famafrench import FamaFrenchReader
from pandas_datareader.fred import FredReader
from pandas_datareader.oecd import OECDReader

__all__ = [
    "get_data_econdb",
    "get_data_famafrench",
    "get_data_fred",
    "DataReader",
]


def get_data_fred(*args, **kwargs):
    return FredReader(*args, **kwargs).read()


def get_data_famafrench(*args, **kwargs):
    return FamaFrenchReader(*args, **kwargs).read()


def get_data_econdb(*args, **kwargs):
    return EcondbReader(*args, **kwargs).read()


@deprecate_kwarg("access_key", "api_key")
def DataReader(
    name,
    data_source=None,
    start=None,
    end=None,
    retry_count=3,
    pause=0.1,
    session=None,
    api_key=None,
):
    """
    Imports data from a number of online sources.

    Currently supports macroeconomic, central-bank, and factor-oriented
    data sources.

    Parameters
    ----------
    name : str or list of strs
        the name of the dataset. Some data sources (fred) will
        accept a list of names.
    data_source: {str, None}
        the data source ("fred", "famafrench", "oecd", "eurostat", "econdb")
    start : string, int, date, datetime, Timestamp
        left boundary for range (defaults to 1/1/2010)
    end : string, int, date, datetime, Timestamp
        right boundary for range (defaults to today)
    retry_count : {int, 3}
        Number of times to retry query request.
    pause : {numeric, 0.001}
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    api_key : (str, None)
        Optional parameter to specify an API key for certain data sources.

    Examples
    ----------
    # Data from FRED
    vix = DataReader("VIXCLS", "fred")

    # Data from Fama/French
    ff = DataReader("F-F_Research_Data_Factors", "famafrench")
    ff = DataReader("F-F_Research_Data_Factors_weekly", "famafrench")
    ff = DataReader("6_Portfolios_2x3", "famafrench")
    ff = DataReader("F-F_ST_Reversal_Factor", "famafrench")
    """
    expected_source = [
        "bankofcanada",
        "fred",
        "famafrench",
        "oecd",
        "eurostat",
        "econdb",
    ]

    if data_source not in expected_source:
        msg = "data_source=%r is not implemented" % data_source
        raise NotImplementedError(msg)

    if data_source == "bankofcanada":
        return BankOfCanadaReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    if data_source == "fred":
        return FredReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    if data_source == "famafrench":
        return FamaFrenchReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    if data_source == "oecd":
        return OECDReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()
    if data_source == "eurostat":
        return EurostatReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()
    if data_source == "econdb":
        return EcondbReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    msg = "data_source=%r is not implemented" % data_source
    raise NotImplementedError(msg)


def Options(symbol, data_source=None, session=None):
    if data_source is None:
        warnings.warn(
            "Options(symbol) is deprecated and no longer part of the public macro/factor API.",
            FutureWarning,
            stacklevel=2,
        )
    raise ImmediateDeprecationError(DEP_ERROR_MSG.format("Options"))



from pandas import DataFrame
import pytest

import pandas_datareader as pdr
from pandas_datareader.data import DataReader

pytestmark = pytest.mark.stable


class TestDataReader:

    def test_read_fred(self):
        vix = DataReader("VIXCLS", "fred")
        assert isinstance(vix, DataFrame)

    def test_not_implemented(self):
        with pytest.raises(NotImplementedError):
            DataReader("NA", "NA")

    @pytest.mark.parametrize(
        "source",
        [
            "yahoo",
            "iex",
            "stooq",
            "quandl",
            "moex",
            "tiingo",
            "naver",
            "av-daily",
        ],
    )
    def test_non_macro_routes_not_implemented(self, source):
        with pytest.raises(NotImplementedError):
            DataReader("AAPL", source)

    def test_top_level_exports_focus_on_macro_and_factor_sources(self):
        assert hasattr(pdr, "get_data_fred")
        assert hasattr(pdr, "get_data_famafrench")
        assert hasattr(pdr, "read_macro")
        assert not hasattr(pdr, "get_data_yahoo")
        assert not hasattr(pdr, "get_data_tiingo")
        assert not hasattr(pdr, "get_quote_yahoo")



from functools import wraps
from io import StringIO
import sys
from urllib.error import HTTPError
import warnings

from pandas.api.types import is_list_like, is_number
from pandas.io import common as com
from pandas.testing import assert_frame_equal

__all__ = [
    "HTTPError",
    "StringIO",
    "deprecate_kwarg",
    "get_filepath_or_buffer",
    "assert_frame_equal",
    "is_list_like",
    "is_number",
    "PYTHON_LT_3_10",
]


def deprecate_kwarg(old_arg_name, new_arg_name, mapping=None, stacklevel=2):
    """
    Decorator to deprecate a keyword argument of a function.

    Vendored from pandas.util._decorators to avoid relying on a private pandas
    API removed in pandas 3.0 (pandas-datareader issue #1005).

    Parameters
    ----------
    old_arg_name : str
        Name of argument to deprecate.
    new_arg_name : str or None
        Replacement argument name. Pass None to warn without renaming.
    mapping : dict or callable, optional
        Translate old argument values to new ones.
    stacklevel : int, default 2
        Passed to warnings.warn.
    """
    if mapping is not None and not hasattr(mapping, "get") and not callable(mapping):
        raise TypeError(
            "mapping from old to new argument values must be dict or callable!"
        )

    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_arg_value = kwargs.pop(old_arg_name, None)

            if old_arg_value is not None:
                if new_arg_name is None:
                    msg = (
                        f"the {repr(old_arg_name)} keyword is deprecated and "
                        "will be removed in a future version. Please take "
                        f"steps to stop the use of {repr(old_arg_name)}"
                    )
                    warnings.warn(msg, FutureWarning, stacklevel=stacklevel)
                    kwargs[old_arg_name] = old_arg_value
                    return func(*args, **kwargs)
                elif mapping is not None:
                    if callable(mapping):
                        new_arg_value = mapping(old_arg_value)
                    else:
                        new_arg_value = mapping.get(old_arg_value, old_arg_value)
                    msg = (
                        f"the {old_arg_name}={repr(old_arg_value)} keyword is "
                        f"deprecated, use {new_arg_name}={repr(new_arg_value)} instead."
                    )
                else:
                    new_arg_value = old_arg_value
                    msg = (
                        f"the {repr(old_arg_name)} keyword is deprecated, "
                        f"use {repr(new_arg_name)} instead."
                    )

                warnings.warn(msg, FutureWarning, stacklevel=stacklevel)
                if kwargs.get(new_arg_name) is not None:
                    msg = (
                        f"Can only specify {repr(old_arg_name)} "
                        f"or {repr(new_arg_name)}, not both."
                    )
                    raise TypeError(msg)
                kwargs[new_arg_name] = new_arg_value
            return func(*args, **kwargs)

        return wrapper

    return _decorator


def get_filepath_or_buffer(filepath_or_buffer, encoding=None, compression=None):
    # Dictionaries are no longer considered valid inputs
    # for "get_filepath_or_buffer" starting in pandas >= 0.20.0
    if isinstance(filepath_or_buffer, dict):
        return filepath_or_buffer, encoding, compression
    try:
        tmp = com._get_filepath_or_buffer(
            filepath_or_buffer, encoding=encoding, compression=None
        )
        return tmp.filepath_or_buffer, tmp.encoding, tmp.compression
    except AttributeError:
        tmp = com.get_filepath_or_buffer(
            filepath_or_buffer, encoding=encoding, compression=None
        )
        return tmp


PYTHON_LT_3_10 = sys.version_info <= (3, 10)
